diff mbox

C++ PATCH for c++/56291 (ICE with constexpr constructor)

Message ID 511A7D81.1090101@redhat.com
State New
Headers show

Commit Message

Jason Merrill Feb. 12, 2013, 5:36 p.m. UTC
When a class has all non-polymorphic bases but has virtual functions 
itself, its vtable pointer goes before the bases but is still 
initialized after them, so sort_constexpr_mem_initializers needs to 
handle this case as well.

Tested x86_64-pc-linux-gnu, applying to trunk and 4.7.
diff mbox

Patch

commit 1e98e1cbcfd70b4ef013c3d8f046e9576d19e1ba
Author: Jason Merrill <jason@redhat.com>
Date:   Tue Feb 12 12:02:58 2013 -0500

    	PR c++/56291
    	* semantics.c (sort_constexpr_mem_initializers): Handle
    	vptr out of order.

diff --git a/gcc/testsuite/g++.dg/cpp0x/constexpr-virtual4.C b/gcc/testsuite/g++.dg/cpp0x/constexpr-virtual4.C
new file mode 100644
index 0000000..32cee96
--- /dev/null
+++ b/gcc/testsuite/g++.dg/cpp0x/constexpr-virtual4.C
@@ -0,0 +1,18 @@ 
+// PR c++/56291
+// { dg-options -std=c++11 }
+
+class Base
+{
+public:
+ constexpr Base() : v(1) {};
+ int v;
+};
+
+class Derived : public Base
+{
+public:
+ constexpr Derived() : Base() {};
+ virtual void function();
+};
+
+Derived d;
diff --git a/gcc/cp/semantics.c b/gcc/cp/semantics.c
index e3dea09..e31ae30 100644
--- a/gcc/cp/semantics.c
+++ b/gcc/cp/semantics.c
@@ -5948,31 +5948,39 @@  check_constexpr_ctor_body (tree last, tree list)
 /* V is a vector of constructor elements built up for the base and member
    initializers of a constructor for TYPE.  They need to be in increasing
    offset order, which they might not be yet if TYPE has a primary base
-   which is not first in the base-clause.  */
+   which is not first in the base-clause or a vptr and at least one base
+   all of which are non-primary.  */
 
 static vec<constructor_elt, va_gc> *
 sort_constexpr_mem_initializers (tree type, vec<constructor_elt, va_gc> *v)
 {
   tree pri = CLASSTYPE_PRIMARY_BINFO (type);
+  tree field_type;
   constructor_elt elt;
   int i;
 
-  if (pri == NULL_TREE
-      || pri == BINFO_BASE_BINFO (TYPE_BINFO (type), 0))
+  if (pri)
+    field_type = BINFO_TYPE (pri);
+  else if (TYPE_CONTAINS_VPTR_P (type))
+    field_type = vtbl_ptr_type_node;
+  else
     return v;
 
-  /* Find the element for the primary base and move it to the beginning of
-     the vec.  */
+  /* Find the element for the primary base or vptr and move it to the
+     beginning of the vec.  */
   vec<constructor_elt, va_gc> &vref = *v;
-  pri = BINFO_TYPE (pri);
-  for (i = 1; ; ++i)
-    if (TREE_TYPE (vref[i].index) == pri)
+  for (i = 0; ; ++i)
+    if (TREE_TYPE (vref[i].index) == field_type)
       break;
 
-  elt = vref[i];
-  for (; i > 0; --i)
-    vref[i] = vref[i-1];
-  vref[0] = elt;
+  if (i > 0)
+    {
+      elt = vref[i];
+      for (; i > 0; --i)
+	vref[i] = vref[i-1];
+      vref[0] = elt;
+    }
+
   return v;
 }
 